###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import json
import readchar
import requests
import textwrap
import subprocess
import hashlib
import sys
import pandas as pd

from tabulate import tabulate
from cmflib.cli.utils import find_root
from cmflib.utils.dvc_config import DvcConfig
from cmflib.cmf_exception_handling import CmfNotConfigured

def is_url(url)-> bool:
    from urllib.parse import urlparse
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def is_git_repo():
    git_dir = os.path.join(os.getcwd(), '.git')
    print("git_dir", git_dir)
    result = os.path.exists(git_dir) and os.path.isdir(git_dir)
    if result:
        return f"A Git repository already exists in {git_dir}."
    else:
        return


def get_python_env(env_name='cmf'):
    # Determine the type of Python environment and return its details
    try:
        # Check if the environment is a Conda environment
        if os.getenv('CONDA_PREFIX') is not None:  # If Conda is activated

            # Step 1: Get the list of conda packages
            conda_packages = subprocess.check_output(['conda', 'list', '--export']).decode('utf-8').splitlines()

            # Step 2: Get the list of pip packages
            pip_packages = subprocess.check_output(['pip', 'freeze']).decode('utf-8').splitlines()

            # Step 3: Get the list of channels from the current conda environment
            channels_raw = subprocess.check_output(['conda', 'config', '--show', 'channels']).decode('utf-8').splitlines()

            # Filter out lines that start with 'channels:' and any empty or commented lines
            channels = [line.strip().lstrip('- ').strip() for line in channels_raw if line and not line.startswith('channels:') and not line.startswith('#')]

            # Step 4: Create a YAML structure for the environment
            env_data = {
                'name': env_name,  # Name the environment -- don't provide the name 
                'channels': channels,  # Add the cleaned channels list
                'dependencies': [],
            }

            # Add conda packages to dependencies
            for package in conda_packages:
                if not package.startswith('#') and len(package.strip()) > 0:
                    env_data['dependencies'].append(package)

            # Add pip packages under a pip section in dependencies
            if pip_packages:
                pip_section = {'pip': pip_packages}
                env_data['dependencies'].append(pip_section)

            return env_data

        else:
            # If not conda, assume virtualenv/pip
            # Step 1: Get the list of pip packages
            pip_packages = subprocess.check_output(['pip', 'freeze']).decode('utf-8').splitlines()

            return pip_packages

    except Exception as e:
        print(f"An error occurred: {e}")

    return


def get_md5_hash(output):
    import hashlib

    # Convert the string to bytes (utf-8 encoding)
    byte_content = output.encode('utf-8')

    # Create an MD5 hash object
    md5_hash = hashlib.md5()

    # Update the hash with the byte content
    md5_hash.update(byte_content)

    # Return the hexadecimal digest
    hash_for_op = md5_hash.hexdigest()

    return hash_for_op
        

def change_dir(cmf_init_path):
    logging_dir = os.getcwd()
    if not logging_dir == cmf_init_path:
        os.chdir(cmf_init_path)
    return logging_dir

def list_conda_packages_json() -> list:
    """Return a list of installed Conda packages and their versions."""
    try:
        result = subprocess.run(['conda', 'list', '--json'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        return json.loads(result.stdout)
    except (subprocess.CalledProcessError, json.JSONDecodeError):
        return []


# Generate SciToken dynamically 
def generate_osdf_token(key_id, key_path, key_issuer) -> str:

    #for SciToken Generation & Validation
    # Error: Skipping analyzing "scitokens": module is installed, but missing library stubs or py.typed marker
    import scitokens    # type: ignore
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.backends import default_backend

    dynamic_pass="" #Initialize Blank dynamic Password

    #Read Private Key using load_pem_private_key() method 
    if not os.path.exists(key_path):
        print(f"File {key_path} does not exist.")
        return dynamic_pass

    try:
        with open(key_path, "r") as file_pointer:
            private_key_contents = file_pointer.read()

        loaded_private_key = serialization.load_pem_private_key(
            private_key_contents.encode(),
            password=None, # Assumes Private key. Update this if password is used for Private Key
            backend=default_backend()
        )

        if is_url(key_issuer):
            token = scitokens.SciToken(key=loaded_private_key, key_id=key_id) #Generate SciToken
            #token.update_claims({"iss": key_issuer, "scope": "write:/ read:/", "aud": "NRP", "sub": "NRP"})
            token.update_claims({"scope": "write:/ read:/", "aud": "NRP", "sub": "NRP"}) #TODO: Figure out how to supply these as input params

            # Serialize the token to a string
            token_ser = token.serialize(issuer=key_issuer) 
            #Key_issuer is something like ""https://t.nationalresearchplatform.org/fdp"

            #Stringify token_str
            token_str=token_ser.decode()
            dynamic_pass="Bearer "+ token_str
        else:
            print(f"{key_issuer} is not a valid URL.")

    except Exception as err:
        print(f"Unexpected {err}, {type(err)}")

    return dynamic_pass


def branch_exists(repo_owner: str, repo_name: str, branch_name: str) -> bool:
    """
    Check if a branch exists in a GitHub repository.

    Args:
        repo_owner: The owner of the GitHub repository.
        repo_name: The name of the GitHub repository.
        branch_name: The name of the branch to check.

    Returns:
        bool: True if the branch exists, otherwise False.
    """
    url = f"https://api.github.com/repos/{repo_owner}/{repo_name}/branches/{branch_name}"
    res = requests.get(url)

    if res.status_code == 200:
        return True
    return False


def display_table(df: pd.DataFrame, columns: list) -> None:
    """
    Display the DataFrame in a paginated table format with text wrapping for better readability.
    Parameters:
    - df: The DataFrame to display.
    - columns: The columns to display in the table.
    """
    # Rearranging columns
    df = df[columns]
    df = df.copy()
    
    # Wrap text in object-type columns to a width of 14 characters.
    # This ensures that long strings are displayed neatly within the table.
    for col in df.select_dtypes(include=["object"]).columns:
        df[col] = df[col].apply(lambda x: textwrap.fill(x, width=14) if isinstance(x, str) else x)

    total_records = len(df)
    start_index = 0  

    # Display up to 20 records per page for better readability. 
    # This avoids overwhelming the user with too much data at once, especially for larger mlmd files.
    while True:
        end_index = start_index + 20
        records_per_page = df.iloc[start_index:end_index]
        
        # Display the table.
        # Fix mypy warning by converting DataFrame to list of lists and columns to list of strings.
        table = tabulate(
            records_per_page.values.tolist(),  # Convert DataFrame to list of lists for tabulate
            headers=[str(col) for col in df.columns],  # Ensure headers are list of strings
            tablefmt="grid",
            showindex=False,
        )
        print(table)

        # Check if we've reached the end of the records.
        if end_index >= total_records:
            print("\nEnd of records.")
            break

        # Ask the user for input to navigate pages.
        print("Press any key to see more or 'q' to quit: ", end="", flush=True)
        user_input = readchar.readchar()
        if user_input.lower() == 'q':
            break
        
        # Update start index for the next page.
        start_index = end_index 


def fetch_cmf_config_path() -> tuple[dict, str]: 
    """ Fetches the CMF configuration and its file path.
    Returns: 
        tuple[dict, str]: A tuple containing the DVC configuration (dict) and the CMF config file path (str).
    Raises: 
        CmfNotConfigured: If the CMF configuration is missing or not properly set up.
    """
    # User can provide different name for cmf configuration file using CONFIG_FILE environment variable.
    # If CONFIG_FILE is not provided, default file name is .cmfconfig
    cmf_config = os.environ.get("CONFIG_FILE", ".cmfconfig") 
    error_message = "'cmf' is not configured.\nExecute 'cmf init' command."
    
    # Fetch DVC configuration 
    dvc_output = DvcConfig.get_dvc_config() 
    if not isinstance(dvc_output, dict): 
        raise CmfNotConfigured(error_message)
    
    # Find the root directory containing the CMF config file 
    cmf_config_root = find_root(cmf_config) 
    if "'cmf' is not configured" in cmf_config_root: 
        raise CmfNotConfigured(error_message)
    
    # Construct the full path to the configuration file 
    config_file_path = os.path.join(cmf_config_root, cmf_config)
    return dvc_output, config_file_path


def get_postgres_config() -> dict:
    """
    Get PostgreSQL configuration from environment variables.

    Returns:
        dict: A dictionary containing PostgreSQL configuration.
    """
    IP = os.getenv('MYIP')
    HOSTNAME = os.getenv('HOSTNAME')
    HOST = ""
    if(HOSTNAME!="localhost"):
        HOST = HOSTNAME if HOSTNAME else ""
    else:
        HOST = IP if IP else ""
    POSTGRES_DB = os.getenv('POSTGRES_DB')
    POSTGRES_USER = os.getenv('POSTGRES_USER')
    POSTGRES_PASSWORD = os.getenv('POSTGRES_PASSWORD')
    config_dict = {"host":HOST, "port":"5432", "user": POSTGRES_USER, "password": POSTGRES_PASSWORD, "dbname": POSTGRES_DB}
    #print("config_dict = ", config_dict)
    return config_dict


def calculate_md5(file_path):
    """
    Calculate MD5 hash for a file
    
    Args:
        file_path (str): Path to the file
        
    Returns:
        str: MD5 hash of the file
    """
    # Check if file exists
    if not os.path.isfile(file_path):
        print(f"Error: File '{file_path}' not found.")
        sys.exit(1)
        
    # Calculate MD5 hash
    md5_hash = hashlib.md5()
    
    # Read file in chunks to handle large files efficiently
    with open(file_path, 'rb') as file:
        # Read in 4MB chunks
        for chunk in iter(lambda: file.read(4096 * 1024), b''):
            md5_hash.update(chunk)
            
    # Return the hexadecimal digest
    return md5_hash.hexdigest()



###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse

from cmflib.cli.command import CmdBase
from cmflib.utils.helper_functions import branch_exists
from cmflib.commands.artifact.pull import CmdArtifactPull
from cmflib.commands.metadata.pull import CmdMetadataPull
from cmflib.cmf_exception_handling import MsgSuccess, MsgFailure
from cmflib.dvc_wrapper import git_get_repo, git_get_pull, git_get_branch


class CmdRepoPull(CmdBase):
    def git_pull(self):
        # Getting GitHub URL from cmf init command
        url = git_get_repo()
        # Example url = https://github.com/ABC/my-repo OR https://github.com/ABC/my-repo.git
        # Check if the URL is a GitHub URL
        if "github.com" not in url:
            raise MsgFailure(msg_str="The repository URL is not a GitHub URL.")
        
        # Extracting the repository owner and name from the URL
        # repo_owner = ABC, repo_name = my-repo
        url_parts = url.split("/")
        repo_owner = url_parts[-2]
        repo_name = url_parts[-1].split(".")[0]
        
        # Getting the current branch name
        branch_name = git_get_branch()[0]
        
        # Check whether the branch exists in the GitHub repository
        if branch_exists(repo_owner, repo_name, branch_name):
            # pull the code from mlmd branch
            print("git pull started...")
            stdout, stderr, returncode = git_get_pull(branch_name)
            if returncode != 0:
                raise MsgFailure(msg_str=f"{stderr}")
            return MsgSuccess(msg_str=stdout)
        else:
            raise MsgFailure(msg_str=f"{branch_name} inside {url} does not exist!!")
        
    def run(self, live):
        print("Executing cmf metadata pull command..")
        metadata_pull_instance = CmdMetadataPull(self.args)
        metadata_pull_result = metadata_pull_instance.run(live)
        if metadata_pull_result.status == "success":  
            print(metadata_pull_result.handle())          # Print the message returned by the handle() method of the metadata_pull_result object.
            print("Executing cmf artifact pull command..")
            artifact_pull_instance = CmdArtifactPull(self.args)
            artifact_pull_result = artifact_pull_instance.run(live)
            if artifact_pull_result.status == "success":
                print(artifact_pull_result.handle())   # Print the message returned by the handle() method of the artifact_pull_result object.
                print("Executing git pull command..")
                live.stop()
                return self.git_pull()


def add_parser(subparsers, parent_parser):
    PULL_HELP = "Pull artifacts, metadata files, and source code from the user's artifact repository, cmf-server, and git respectively."

    parser = subparsers.add_parser(
        "pull",
        parents=[parent_parser],
        description="Pull artifacts, metadata files, and source code from the user's artifact repository, cmf-server, and git respectively.",
        help=PULL_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        action="append",
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f", 
        "--file_name", 
        action="append",
        help="Specify output metadata file name.", 
        metavar="<file_name>",
    )

    parser.add_argument(
        "-e",
        "--execution_uuid",
        action="append",
        help="Specify Execution uuid.",
        metavar="<exec_uuid>",
    )

    # The 'artifact_name' parameter is used inside 'cmf artifact pull' command.
    # To avoid errors, it is defined here with a default value of 'None' and hidden from the help text using 'argparse.SUPPRESS'.
    parser.add_argument(
        "-a", 
        "--artifact_name", 
        action="store_const",
        const="None",
        help=argparse.SUPPRESS, 
        metavar="<artifact_name>",
    )

    parser.set_defaults(func=CmdRepoPull)



###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import argparse
import textwrap
import readchar
import pandas as pd

from cmflib import cmfquery
from tabulate import tabulate
from typing import Union, List
from cmflib.cli.command import CmdBase
from cmflib.utils.helper_functions import display_table
from cmflib.cmf_exception_handling import ( 
    PipelineNotFound,
    FileNotFound,
    ArtifactNotFound,
    DuplicateArgumentNotAllowed,
    MissingArgument,
    MsgSuccess
)

class CmdArtifactsList(CmdBase):
    def convert_to_datetime(self, df: pd.DataFrame, col_name: str) -> pd.DataFrame:
        """
        Function to convert a column to datetime format.
        Parameters:
        - df: The DataFrame containing the data.
        - col_name: The name of the column to convert to datetime.
        Returns:
        - The updated DataFrame with the specified column converted to datetime format.
        """  
        # Convert the col_name column to UTC datetime format.
        # The datetime is formatted as "Day DD Mon YYYY HH:MM:SS GMT"
        df=df.copy()
        df[col_name] = pd.to_datetime(df[col_name], unit='ms', utc=True).dt.strftime("%a %d %b %Y %H:%M:%S GMT")
        
        return df
    
    def search_artifact(self, df: pd.DataFrame) -> Union[int, List[int]]:
        """
        Searches for the specified 'artifact_name' in the DataFrame and returns matching IDs.

        Parameters:
        - df: DataFrame to search within.

        Returns:
        - List of matching IDs or -1 if no matches are found.
        """
        # Example of a given sample 'artifact_name' --> artifacts/parsed/train.tsv:12345
        # These are the combinations we are implementing:
        # 1. artifacts/parsed/train.tsv
        # 2. train.tsv   
        # 3. artifacts/parsed/train.tsv:12345
        # 4. train.tsv:12345
        
        # In case of multiple occurrences of 'artifact_name', we need to store the IDs of all matching names.
        # For example, if "metrics" appears multiple times, we store all its IDs.
        matched_ids = []
        artifact_name = self.args.artifact_name[0].strip()
        for index, row in df.iterrows():
            # Extract the base name from the row.
            # eg. artifacts/parsed/train.tsv:12345 --> artifacts/parsed/train.tsv
            name =  row['name'].split(":")[0]      
            if artifact_name == name:             # Match the full path: artifacts/parsed/train.tsv
                matched_ids.append(row['id'])     
            elif artifact_name == name.split('/')[-1]:    # Match only the file name: train.tsv
                matched_ids.append(row['id'])
            elif artifact_name == row['name']:             # Match the full path with hash: artifacts/parsed/train.tsv:12345
                matched_ids.append(row['id'])
            elif artifact_name == row["name"].split('/')[-1]:   # Match only the file name with hash: train.tsv:12345
                matched_ids.append(row['id'])
        
        if len(matched_ids) != 0:
            return matched_ids
        return -1

    def run(self, live):
        cmd_args = {
            "file_name": self.args.file_name,
            "pipeline_name": self.args.pipeline_name,
            "artifact_name": self.args.artifact_name
        }
        for arg_name, arg_value in cmd_args.items():
            if arg_value:
                if arg_value[0] == "":
                    raise MissingArgument(arg_name)
                elif len(arg_value) > 1:
                    raise DuplicateArgumentNotAllowed(arg_name,("-"+arg_name[0]))

        # default path for mlmd file name
        mlmd_file_name = "./mlmd"
        current_directory = os.getcwd()
        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        else:
            mlmd_file_name = self.args.file_name[0].strip()
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
        current_directory = os.path.dirname(mlmd_file_name)
        if not os.path.exists(mlmd_file_name):
            raise FileNotFound(mlmd_file_name, current_directory)
        
        # Creating cmfquery object.
        query = cmfquery.CmfQuery(mlmd_file_name)
        
        # Check if pipeline exists in mlmd.
        pipeline_name = self.args.pipeline_name[0]
        df = query.get_all_artifacts_by_context(pipeline_name)

        if df.empty:
            raise PipelineNotFound(pipeline_name)
        else:
            if not self.args.artifact_name:         # If self.args.artifact_name is None or an empty list ([]). 
                pass
            else:
                artifact_ids = self.search_artifact(df)
                if(artifact_ids != -1):
                    # Multiple/Single artifact names exist with the same name.
                    for artifact_id in artifact_ids:
                        # Filter the DataFrame to retrieve rows corresponding to the current ID.
                        filtered_data = df.loc[df['id'] == artifact_id] 

                        # Converting "create_time_since_epoch" and "last_update_time_since_epoch" to datetime format.
                        filtered_data = self.convert_to_datetime(filtered_data, "create_time_since_epoch")
                        filtered_data = self.convert_to_datetime(filtered_data, "last_update_time_since_epoch")

                        # Rearranging columns: Start with fixed columns and appending the remaining columns.
                        updated_columns = ["id", "name", "type", "create_time_since_epoch", "url", "Commit", "uri", "last_update_time_since_epoch"] 
                        updated_columns += [ col for col in filtered_data.columns if col not in updated_columns]
                        
                        filtered_data = filtered_data[updated_columns]

                        # Drop columns that start with 'custom_properties_' and that contains NaN values
                        columns_to_drop = [col for col in filtered_data.columns if col.startswith('custom_properties_') and df[col].isna().any()]
                        filtered_data = filtered_data.drop(columns=columns_to_drop)

                        # Wrap text in object-type columns to a width of 30 characters for better readability.
                        for col in filtered_data.select_dtypes(include=['object']).columns:
                            filtered_data[col] = filtered_data[col].apply(lambda x: textwrap.fill(x, width=30) if isinstance(x, str) else x)
                        
                        # For a single artifact name, display the table in a horizontal format:
                        # Set 'id' as the index.
                        filtered_data.set_index("id", inplace=True)
                        # Transpose the DataFrame to make rows into columns.
                        filtered_data = filtered_data.T.reset_index()
                        # Rename the first column back to 'id' for consistency.
                        filtered_data.columns.values[0] = 'id'
                        
                        # Display the formatted and transposed table using the 'tabulate' library.
                        table = tabulate(
                            filtered_data,
                            headers=filtered_data.columns,  # Use column names as headers.
                            tablefmt="grid",                # Use grid format for table borders.
                            showindex=False,                # Do not display the default index.
                        )
                        print(table)
                        print()

                        print("Press any key to see more or 'q' to quit: ", end="", flush=True)
                        user_input = readchar.readchar()
                        if user_input.lower() == 'q':
                            break
                    return MsgSuccess(msg_str = "End of records..")
                else:
                    raise ArtifactNotFound(self.args.artifact_name[0])
        
        df = self.convert_to_datetime(df, "create_time_since_epoch")
        display_table(df, ["id", "name", "type", "create_time_since_epoch", "url", "Commit", "uri"])

        return MsgSuccess(msg_str = "Done.")


def add_parser(subparsers, parent_parser):
    ARTIFACT_LIST_HELP = "Displays artifacts from the input metadata file with a few properties in a 7-column table, limited to 20 records per page."

    parser = subparsers.add_parser(
        "list",
        parents=[parent_parser],
        description="Displays artifacts from the input metadata file with a few properties in a 7-column table, limited to 20 records per page.",
        help=ARTIFACT_LIST_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    required_argumets = parser.add_argument_group("required arguments")

    required_argumets.add_argument(
        "-p", 
        "--pipeline_name", 
        required=True,
        action="append",
        help="Specify pipeline name.", 
        metavar="<pipeline_name>", 
    )

    parser.add_argument(
        "-f", 
        "--file_name",
        action="append",
        help="Specify input metadata file name.",
        metavar="<file_name>",
    )

    parser.add_argument(
        "-a", 
        "--artifact_name", 
        action="append",
        help="Specify the artifact name to display detailed information about the given artifact name.",
        metavar="<artifact_name>",
    )

    parser.set_defaults(func=CmdArtifactsList)


###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import os
import argparse

from cmflib import cmfquery
from cmflib.storage_backends import (
    minio_artifacts,
    local_artifacts,
    amazonS3_artifacts,
    sshremote_artifacts,
    osdf_artifacts,
)
from cmflib.cli.command import CmdBase
from cmflib.cmf_exception_handling import (
    PipelineNotFound, 
    FileNotFound, 
    MissingArgument,
    ExecutionsNotFound, 
    ArtifactNotFound, 
    BatchDownloadFailure, 
    BatchDownloadSuccess,
    ObjectDownloadFailure, 
    ObjectDownloadSuccess,
    DuplicateArgumentNotAllowed,
    MsgSuccess,
    MsgFailure
)
from cmflib.utils.helper_functions import fetch_cmf_config_path

class CmdArtifactPull(CmdBase):
    def split_url_pipeline(self, url: str, pipeline_name: str):
       # This function takes url and pipeline_name as a input parameter
       # return string which contains the artifact repo path of the artifact
       # url = Test-env:/home/user/local-storage/files/md5/23/6d9502e0283d91f689d7038b8508a2
       # pipeline_name = Test-env

       # checking whether pipeline name exist inside url
       if pipeline_name in url:
            # if multiple pipelines logs same artifact, then spliting them using ',' delimiter 
            if "," in url:
                urls = url.split(",")
                # iterate over each urls
                for u in urls:
                    # assign u to url if pipeline name exist
                    if pipeline_name in u:
                        url = u
            # splitting url using ':' delimiter 
            # token = ["Test-env","home/user/local-storage/files/md5/23/6d9502e508a2"]
            token = url.split(":")
            # removing 1st element from token i.e pipeline name
            # output token will be ["home/user/local-storage/files/md5/23/6d9502e508a2"]
            token.pop(0)
            if len(token) > 1:
                # in case of metrics we have multiple ':' in its url
                # concating remaining tokens after removing pipeline_name using ':' delimiter
                token_str = ":".join(token)
                return token_str
            return "".join(token)

    def extract_repo_args(self, type: str, name: str, url: str, current_directory: str):
        # Extracting the repository URL, current path, bucket name, and other relevant
        # information from the user-supplied arguments.
        # url = Test-env:/home/user/local-storage/files/md5/06/d100ff3e04e2c87bf20f0feacc9034,
        #          Second-env:/home/user/local-storage/files/md5/06/d100ff3e04e2c"
        # s_url = Url without pipeline name
        s_url = self.split_url_pipeline(url, self.args.pipeline_name[0])

        # got url in the form of /home/user/local-storage/files/md5/06/d100ff3e04e2c
        # spliting url using '/' delimiter
        token = s_url.split("/")

        # name = artifacts/model/model.pkl
        name = name.split(":")[0]
        if type == "minio":
            token_length = len(token)

            # assigned 2nd position element to bucket_name
            bucket_name = token[2]

            # The folder structure of artifact data has been updated due to a change in the DVC 3.0 version
            # Previously, the structure was dvc-art/23/69v2uu3jeejjeiw
            # but now it includes additional directories and has become files dvc-art/files/md5/23/69v2uu3jeejjeiw.
            # Consequently, the previous logic takes only the last 2 elements from the list of tokens,
            # but with the new structure, it needs to take the last 4 elements.

            # get last 4 element inside token
            token = token[(token_length-4):]

            # join last 4 token using '/' delimiter
            object_name = "/".join(token)
            # output = files/md5/23/69v2uu3jeejjeiw

            path_name = current_directory + "/" + name
            return bucket_name, object_name, path_name

        elif type == "local":
            token_length = len(token)
            download_loc = current_directory + "/" + name
            # local artifact repo path =  local-storage/files/md5/23/69v2uu3jeejjeiw.
            # token is a list = ['local-storage', 'files', 'md5', '23', '69v2uu3jeejjeiw']
            # get last 4 element inside token
            token = token[(token_length-4):]
            # join last 4 token using '/' delimiter
            current_dvc_loc = "/".join(token)
            return current_dvc_loc, download_loc

        elif type == "ssh":
            # token = ['ssh:', '', 'XX.XX.XX.XX:22', 'home', 'user', 'ssh-storage', 'files', 'md5', '23', '6d9502e0283d91f689d7038b8508a2']
            host_with_port = token[2].split(":")
            host = host_with_port[0]
            # Update token list by removing the first three items
            # token = ['home', 'user', 'ssh-storage', 'files', 'md5', '23', '6d9502e0283d91f689d7038b8508a2']
            current_loc = '/' + '/'.join(token[3:])
            return host, current_loc, name
        
        elif type == "osdf":
            token_length = len(token)
            download_loc = current_directory + "/" + name if current_directory != ""  else name
            #current_dvc_loc = (token[(token_length - 2)] + "/" + token[(token_length - 1)])
            #return FQDNL of where to download from, where to download to, what the artifact will be named
            return s_url, download_loc, name
        else:
            # sometimes s_url is empty - this shouldn't happen technically
            # sometimes s_url is not starting with s3:// - technically this shouldn't happen
            if s_url and s_url.startswith("s3://"):
                url_with_bucket = s_url.split("s3://")[1]
                # url_with_bucket = mybucket/user/files/md5/23/6d9502e0283d91f689d7038b8508a2
                # splitting the string using '/' as the delimiter
                # bucket_name = mybucket
                # object_name = user/files/md5/23/6d9502e0283d91f689d7038b8508a2
                bucket_name, object_name = url_with_bucket.split('/', 1)
                download_loc =  current_directory + "/" + name if current_directory != ""  else name
                #print(download_loc)
                return bucket_name, object_name, download_loc
            else:
                # returning bucket_name, object_name and download_loc returning as empty
                return "", "", ""

    def search_artifact(self, input_dict, remote):
        flag = True
        artifact_name = self.args.artifact_name[0]
        
        # This function takes input_dict as input artifact
        for name, url in input_dict.items():
            if not isinstance(url, str):
                continue
            # Splitting the 'name' using ':' as the delimiter and storing the first argument in the 'file_path' variable.
            # eg name = ./a/data.xml.gz:12345abcd -->  a/data.xml.gz
            file_path = name.split(":")[0]
            # Splitting the path on '/' to extract the file name, excluding the directory structure.
            # eg name = ./a/data.xml.gz --> data.xml.gz
            file_name = file_path.split('/')[-1]
           
            if remote == "osdf":
                # Extracting the artifact hash from the artifact name.
                artifact_hash = name = name.split(":")[1]
                return name, url, artifact_hash
            elif name == artifact_name or file_path == artifact_name or file_name == artifact_name:
                # If the artifact name matches, set flag to False and break the loop
                flag = False
                break
            
        if flag:
            # Raise an exception if the artifact is not found
            raise ArtifactNotFound(artifact_name)
        return name, url

    def run(self, live):
        output, config_file_path = fetch_cmf_config_path()
        
        cmd_args = {
            "file_name": self.args.file_name,
            "pipeline_name": self.args.pipeline_name,
            "artifact_name": self.args.artifact_name
        }
        for arg_name, arg_value in cmd_args.items():
            if arg_value:
                if arg_value[0] == "":
                    raise MissingArgument(arg_name)
                elif len(arg_value) > 1:
                    raise DuplicateArgumentNotAllowed(arg_name,("-"+arg_name[0]))
                
        # check whether 'mlmd' file exist in current directory 
        # or in the directory provided by user
        # pipeline_name = self.args.pipeline_name
        current_directory = os.getcwd()
        mlmd_file_name = "./mlmd"
        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        else:
            mlmd_file_name = self.args.file_name[0].strip()
            if "/" not in mlmd_file_name:
                mlmd_file_name = "./"+mlmd_file_name
        current_directory = os.path.dirname(mlmd_file_name)

        if not self.args.artifact_name:         # If self.args.artifact_name[0] is None or an empty list ([]). 
            pass
        
        if not os.path.exists(mlmd_file_name):   #checking if MLMD files exists
            raise FileNotFound(mlmd_file_name, current_directory)
        query = cmfquery.CmfQuery(mlmd_file_name)
        
        pipeline_name = self.args.pipeline_name[0]
        if not pipeline_name in query.get_pipeline_names():   #checking if pipeline name exists in mlmd
            raise PipelineNotFound(pipeline_name)
        
        # getting all pipeline stages[i.e Prepare, Featurize, Train and Evaluate]
        stages = query.get_pipeline_stages(pipeline_name)
        executions = []
        identifiers = []
        for stage in stages:
            # getting all executions for stages
            executions = query.get_all_executions_in_stage(stage)
            # check if stage has executions
            if len(executions) > 0:
                 # converting it to dictionary
                dict_executions = executions.to_dict("dict")
                # append id's of executions inside identifiers
                for id in dict_executions["id"].values():
                    identifiers.append(id)
            else:
                print("No Executions found for " + stage + " stage.")
        # created dictionary
        name_url_dict = {}
        if len(identifiers) == 0:  # check if there are no executions
            raise ExecutionsNotFound()
        for identifier in identifiers:
            get_artifacts = query.get_all_artifacts_for_execution(
                identifier
            )  # getting all artifacts with id
            # skipping artifacts if it is type of label
            temp_dict = {
                name: url
                for name, url, artifact_type in zip(get_artifacts['name'], get_artifacts['url'], get_artifacts['type'])
                if artifact_type != "Label"
            }
            name_url_dict.update(temp_dict) # updating name_url_dict with temp_dict
        # name_url_dict = ('artifacts/parsed/test.tsv:6f597d341ceb7d8fbbe88859a892ef81', 'Test-env:/home/sharvark/local-storage/6f/597d341ceb7d8fbbe88859a892ef81'
        # name_url_dict = ('artifacts/parsed/test.tsv:6f597d341ceb7d8fbbe88859a892ef81', 'Test-env:/home/sharvark/local-storage/6f/597d341ceb7d8fbbe88859a892ef81,Second-env:/home/sharvark/local-storage/6f/597d341ceb7d8fbbe88859a892ef81')
        """
           There are multiple scenarios for cmf artifact pull 
           Code checks if self.args.artifact_name[0] is provided by user or not
           under these conditions there are two more conditions
              1. if file is not .dir (single file) 
                   Download single file
              2. else file is .dir (directory)
                   download all files from directory
                     
        """
        dvc_config_op = output
        if dvc_config_op["core.remote"] == "minio":
            minio_class_obj = minio_artifacts.MinioArtifacts(dvc_config_op)
            # Check if a specific artifact name is provided as input.
            if self.args.artifact_name: 
                # Search for the artifact in the metadata store.
                # If the artifact is not found, an error will be raised automatically.
                output = self.search_artifact(name_url_dict, dvc_config_op["core.remote"])
                # output[0] = artifact_name
                # output[1] = url
                # output[2] = hash
                # Extract repository arguments specific to MinIO.
                minio_args = self.extract_repo_args("minio", output[0], output[1], current_directory)

                # Check if the object name doesn't end with `.dir` (indicating it's a file).
                if not minio_args[1].endswith(".dir"):
                    # Download a single file from MinIO.
                    object_name, download_loc, download_flag = minio_class_obj.download_file(
                        current_directory,
                        minio_args[0], # bucket_name
                        minio_args[1], # object_name
                        minio_args[2], # path_name
                    )
                    if download_flag:
                        # Return success if the file is downloaded successfully.
                        return ObjectDownloadSuccess(object_name, download_loc)
                    raise ObjectDownloadFailure(object_name)
                else:
                    # If object name ends with `.dir`, download multiple files from a directory 
                    # return total_files_in_directory, files_downloaded
                    total_files_in_directory, dir_files_downloaded, download_flag = minio_class_obj.download_directory(
                        current_directory,
                        minio_args[0], # bucket_name
                        minio_args[1], # object_name
                        minio_args[2], # path_name
                    )
        
                    if download_flag:
                        # Return success if all files in the directory are downloaded.
                        return BatchDownloadSuccess(dir_files_downloaded)
                    # Calculate the number of files that failed to download.
                    file_failed_to_download = total_files_in_directory - dir_files_downloaded
                    raise BatchDownloadFailure(dir_files_downloaded, file_failed_to_download)
            
            else:
                # Handle the case where no specific artifact name is provided.
                files_downloaded = 0
                files_failed_to_download = 0

                # Iterate through the dictionary of artifact names and URLs.
                for name, url in name_url_dict.items():
                    if not isinstance(url, str):    ## Skip invalid URLs.
                        continue
                    minio_args = self.extract_repo_args("minio", name, url, current_directory)

                    # Check if the object name doesn't end with `.dir` (indicating it's a file).
                    if not minio_args[1].endswith(".dir"):
                        # Download a single file from MinIO.
                        object_name, download_loc, download_flag = minio_class_obj.download_file(
                            current_directory,
                            minio_args[0], # bucket_name
                            minio_args[1], # object_name
                            minio_args[2], # path_name
                        )

                        # print output here because we are in a loop and can't return the control
                        if download_flag:
                            print(f"object {object_name} downloaded at {download_loc}.")
                            files_downloaded += 1
                        else:
                            print(f"object {object_name} is not downloaded.")
                            files_failed_to_download += 1
                    else:
                        # If object name ends with `.dir`, download multiple files from a directory.
                        total_files_in_directory, dir_files_downloaded, download_flag = minio_class_obj.download_directory(
                            current_directory,
                            minio_args[0], # bucket_name
                            minio_args[1], # object_name
                            minio_args[2], # path_name
                        )
                        # Return success if all files in the directory are downloaded.
                        if download_flag:
                            files_downloaded += dir_files_downloaded
                        else:
                            files_downloaded += dir_files_downloaded
                            files_failed_to_download += (total_files_in_directory - dir_files_downloaded)
                            
                # we are assuming, if files_failed_to_download > 0, it means our download of artifacts is not success
                if not files_failed_to_download:
                    return BatchDownloadSuccess(files_downloaded)
                raise BatchDownloadFailure(files_downloaded, files_failed_to_download)

        elif dvc_config_op["core.remote"] == "local-storage":
            local_class_obj = local_artifacts.LocalArtifacts(dvc_config_op)
            # There are two main conditions
            # Condition 1 - user can use -a paramter for cmf artifact pull command
                # -a can be a dir or a file
            # Condition 2 - user can chose to download all the artifacts in one go. 
                # we can have both dir and files in our list of artifacts
            # Check if a specific artifact name is provided as input.
            if self.args.artifact_name:
                # Search for the artifact in the metadata store.
                # If the artifact is not found, an error will be raised automatically.
                output = self.search_artifact(name_url_dict, dvc_config_op["core.remote"])
                # output[0] = name
                # output[1] = url
                # Extract repository arguments specific to Local repo.
                local_args = self.extract_repo_args("local", output[0], output[1], current_directory)
                # local_args [0] = current_dvc_loc
                # local_args [1] = download_loc
                # Check if the object name doesn't end with `.dir` (indicating it's a file).
                if not local_args[0].endswith(".dir"):
                    # Download a single file from Local.
                    object_name, download_loc, download_flag = local_class_obj.download_file(current_directory, local_args[0], local_args[1])
                    if download_flag:
                        # Return success if the file is downloaded successfully.
                        return ObjectDownloadSuccess(object_name, download_loc)
                    raise ObjectDownloadFailure(object_name)
                    
                else:
                    # If object name ends with `.dir`, download multiple files from a directory 
                    # return total_files_in_directory, files_downloaded
                    total_files_in_directory, dir_files_downloaded, download_flag = local_class_obj.download_directory(current_directory, local_args[0], local_args[1])
        
                    if download_flag:
                        # Return success if all files in the directory are downloaded.
                        return BatchDownloadSuccess(dir_files_downloaded)
                    # Calculate the number of files that failed to download.
                    file_failed_to_download = total_files_in_directory - dir_files_downloaded
                    raise BatchDownloadFailure(dir_files_downloaded, file_failed_to_download)
            else:
                # Handle the case where no specific artifact name is provided.
                files_downloaded = 0
                files_failed_to_download = 0
                # Iterate through the dictionary of artifact names and URLs.
                for name, url in name_url_dict.items():
                    if not isinstance(url, str):
                        continue
                    # name1 - file
                    # name2 - failed file
                    # name3 - dir (5 files)
                    # name4 - dir (4 files) - failed dir - 2 files passed, 2 files failed
                    # name5 - file
                    # name6 - dir - and can't open it (but it has 2 files) .. user don't know 
                    local_args = self.extract_repo_args("local", name, url, current_directory)
                    # local_args [0] = current_dvc_loc
                    # local_args [1] = download_loc
                    # Check if the object name doesn't end with `.dir` (indicating it's a file).
                    if not local_args[0].endswith(".dir"):
                        # Download a single file from Local repo.
                        object_name, download_loc, download_flag = local_class_obj.download_file(
                            current_directory, local_args[0], local_args[1])
                        
                        # print output here because we are in a loop and can't return the control
                        if download_flag:
                            print(f"object {object_name} downloaded at {download_loc}.")
                            files_downloaded += 1
                        else:
                            print(f"object {object_name} is not downloaded.")
                            files_failed_to_download += 1
                    else:
                        # If object name ends with `.dir`, download multiple files from a directory.
                        total_files_in_directory, dir_files_downloaded, download_flag = local_class_obj.download_directory(
                            current_directory, local_args[0], local_args[1])
                        # download_flag is true only when all the files from the directory are successfully downlaoded.
                        if download_flag:
                            files_downloaded += dir_files_downloaded
                        else:
                            files_downloaded += dir_files_downloaded
                            files_failed_to_download += (total_files_in_directory - dir_files_downloaded)
                            
                # we are assuming, if files_failed_to_download > 0, it means our download of artifacts is not success
                if not files_failed_to_download:
                    return BatchDownloadSuccess(files_downloaded)
                raise BatchDownloadFailure(
                            files_downloaded, files_failed_to_download)
                    
        elif dvc_config_op["core.remote"] == "ssh-storage":
            sshremote_class_obj = sshremote_artifacts.SSHremoteArtifacts(dvc_config_op)
            # Check if a specific artifact name is provided as input.
            if self.args.artifact_name:
                # Search for the artifact in the metadata store.
                # If the artifact is not found, an error will be raised automatically.
                output = self.search_artifact(name_url_dict, dvc_config_op["core.remote"])
                # output[0] = name
                # output[1] = url
                # Extract repository arguments specific to ssh-remote.
                args = self.extract_repo_args("ssh", output[0], output[1], current_directory)
                # Check if the object name doesn't end with `.dir` (indicating it's a file).
                if not args[1].endswith(".dir"):
                    # Download a single file from ssh-remote.
                    object_name, download_loc, download_flag = sshremote_class_obj.download_file(
                        args[0], # host,
                        current_directory,
                        args[1], # remote_loc of the artifact
                        args[2]  # name
                    )
                    if download_flag:
                        # Return success if the file is downloaded successfully.
                        return ObjectDownloadSuccess(object_name, download_loc)
                    raise ObjectDownloadFailure(object_name)

                else:
                    # If object name ends with `.dir`, download multiple files from a directory 
                    # return total_files_in_directory, files_downloaded
                    total_files_in_directory, dir_files_downloaded, download_flag = sshremote_class_obj.download_directory(
                        args[0], # host,
                        current_directory,
                        args[1], # remote_loc of the artifact
                        args[2]  # name
                        )
                if download_flag:
                    # Return success if all files in the directory are downloaded.
                    return BatchDownloadSuccess(dir_files_downloaded)
                # Calculate the number of files that failed to download.
                file_failed_to_download = total_files_in_directory - dir_files_downloaded
                raise BatchDownloadFailure(dir_files_downloaded, file_failed_to_download)   
            else:
                # Handle the case where no specific artifact name is provided.
                files_downloaded = 0
                files_failed_to_download = 0
                # Iterate through the dictionary of artifact names and URLs.
                for name, url in name_url_dict.items():
                    if not isinstance(url, str):
                        continue
                    args = self.extract_repo_args("ssh", name, url, current_directory)
                    # Check if the object name doesn't end with `.dir` (indicating it's a file).
                    if not args[1].endswith(".dir"):
                        # Download a single file from ssh-remote.
                        object_name, download_loc, download_flag = sshremote_class_obj.download_file(
                        args[0], # host,
                        current_directory,
                        args[1], # remote_loc of the artifact
                        args[2]  # name
                    )
                        # print output here because we are in a loop and can't return the control
                        if download_flag:
                            print(f"object {object_name} downloaded at {download_loc}.")
                            files_downloaded += 1
                        else:
                            print(f"object {object_name} is not downloaded.")
                            files_failed_to_download += 1
                    else:
                        # If object name ends with `.dir`, download multiple files from a directory.
                        total_files_in_directory, dir_files_downloaded, download_flag = sshremote_class_obj.download_directory(
                            args[0], # host,
                            current_directory,
                            args[1], # remote_loc of the artifact
                            args[2]  # name
                        )
                        if download_flag:
                            files_downloaded += dir_files_downloaded
                        else:
                            files_downloaded += dir_files_downloaded
                            files_failed_to_download += (total_files_in_directory - dir_files_downloaded)
                            
                # we are assuming, if files_failed_to_download > 0, it means our download of artifacts is not success
                if not files_failed_to_download:
                    return BatchDownloadSuccess(files_downloaded)
                raise BatchDownloadFailure(files_downloaded, files_failed_to_download)
        elif dvc_config_op["core.remote"] == "osdf":
            #Regenerate Token for OSDF
            from cmflib.utils.helper_functions import generate_osdf_token
            from cmflib.dvc_wrapper import dvc_add_attribute
            from cmflib.utils.cmf_config import CmfConfig
            #Fetch Config from CMF_Config_File
            cmf_config_file = os.environ.get("CONFIG_FILE", ".cmfconfig")
            cmf_config={}
            cmf_config=CmfConfig.read_config(cmf_config_file)
            #Regenerate password 
            dynamic_password = generate_osdf_token(cmf_config["osdf-key_id"],cmf_config["osdf-key_path"],cmf_config["osdf-key_issuer"])
            #cmf_config["password"]=dynamic_password
            #Update Password in .dvc/config for future use
            dvc_add_attribute(dvc_config_op["core.remote"],"password",dynamic_password)
            #Updating dvc_config_op data structure with new password as well since this is used in download_artifacts() below
            dvc_config_op["remote.osdf.password"]=dynamic_password
            #Need to write to cmfconfig with new credentials
            #CmfConfig.write_config(cmf_config, "osdf", attr_dict, True)
            #Now Ready to do dvc pull 
            cache_path=cmf_config["osdf-cache"]

            osdfremote_class_obj = osdf_artifacts.OSDFremoteArtifacts()
            if self.args.artifact_name:
                # Search for the artifact in the metadata store.
                # If the artifact is not found, an error will be raised automatically.
                output = self.search_artifact(name_url_dict, dvc_config_op["core.remote"])
                # output[0] = name
                # output[1] = url
                # output[3]=artifact_hash
                args = self.extract_repo_args("osdf", output[0], output[1], current_directory)
                download_flag, message = osdfremote_class_obj.download_artifacts(
                    dvc_config_op,
                    args[0], # s_url of the artifact
                    cache_path,
                    current_directory,
                    args[1], # download_loc of the artifact
                    args[2],  # name of the artifact
                    output[3] #Artifact Hash
                )
                
                if download_flag :
                    status = MsgSuccess(msg_str = message)
                else:
                    status = MsgFailure(msg_str = message)
                return status
            else:
                for name, url in name_url_dict.items():
                    total_files_count += 1
                    #print(name, url)
                    if not isinstance(url, str):
                        continue
                    artifact_hash = name.split(':')[1] #Extract Hash of the artifact from name
                    #print(f"Hash for the artifact {name} is {artifact_hash}")
                    args = self.extract_repo_args("osdf", name, url, current_directory)
                        
                    download_flag, message = osdfremote_class_obj.download_artifacts(
                        dvc_config_op,
                        args[0], # host,
                        cache_path,
                        current_directory,
                        args[1], # remote_loc of the artifact
                        args[2],  # name
                        artifact_hash #Artifact Hash
                    )
                    if download_flag:
                        print(message)   #### success message
                        file_downloaded +=1
                    else:
                        print(message)    ### failure message
                Files_failed_to_download = total_files_count - files_downloaded
                if Files_failed_to_download == 0:
                    status = BatchDownloadSuccess(files_downloaded=files_downloaded)
                else:
                    status = BatchDownloadFailure(files_downloaded=files_downloaded, Files_failed_to_download= Files_failed_to_download)
                return status
        elif dvc_config_op["core.remote"] == "amazons3":
            amazonS3_class_obj = amazonS3_artifacts.AmazonS3Artifacts(dvc_config_op)
            if self.args.artifact_name:
                # Search for the artifact in the metadata store.
                # If the artifact is not found, an error will be raised automatically.
                output = self.search_artifact(name_url_dict, dvc_config_op["core.remote"])
                # output[0] = name
                # output[1] = url
                args = self.extract_repo_args("amazons3", output[0], output[1], current_directory)
                if args[0] and args[1] and args[2]:
                    if not args[1].endswith(".dir"):
                        object_name, download_loc, download_flag = amazonS3_class_obj.download_file(
                            current_directory,
                            args[0], # bucket_name
                            args[1], # object_name
                            args[2], # download_loc
                        )
                        if download_flag:
                            return ObjectDownloadSuccess(object_name, download_loc)
                        raise ObjectDownloadFailure(object_name)
                    else:
                        total_files_in_directory, dir_files_downloaded, download_flag = amazonS3_class_obj.download_directory(current_directory,
                            args[0], # bucket_name
                            args[1], # object_name
                            args[2], # download_loc
                            )
                    if download_flag:
                        return BatchDownloadSuccess(dir_files_downloaded)
                    file_failed_to_download = total_files_in_directory - dir_files_downloaded
                    raise BatchDownloadFailure(dir_files_downloaded, file_failed_to_download)
        
            else:
                files_downloaded = 0
                files_failed_to_download = 0
                for name, url in name_url_dict.items():
                    if not isinstance(url, str):
                        continue
                    args = self.extract_repo_args("amazons3", name, url, current_directory)
                    if args[0] and args[1] and args[2]:
                        if not args[1].endswith(".dir"):
                            object_name, download_loc, download_flag = amazonS3_class_obj.download_file(
                                current_directory,
                                args[0], # bucket_name
                                args[1], # object_name
                                args[2], # download_loc
                            )
                            if download_flag:
                                print(f"object {object_name} downloaded at {download_loc}.")
                                files_downloaded += 1
                            else:
                                print(f"object {object_name} is not downloaded.")
                                files_failed_to_download += 1
                        else:
                            total_files_in_directory, dir_files_downloaded, download_flag = amazonS3_class_obj.download_directory(
                            current_directory,
                            args[0], # bucket_name
                            args[1], # object_name
                            args[2], # path_name
                        )
                        # download_flag is true only when all the files from the directory are successfully downlaoded.
                            if download_flag:
                                files_downloaded += dir_files_downloaded
                            else:
                                files_downloaded += dir_files_downloaded
                                files_failed_to_download += (total_files_in_directory - dir_files_downloaded)
                            
                # we are assuming, if files_failed_to_download > 0, it means our download of artifacts is not success
                if not files_failed_to_download:
                    return BatchDownloadSuccess(files_downloaded)
                raise BatchDownloadFailure(files_downloaded, files_failed_to_download)
        else:
            remote = dvc_config_op["core.remote"]
            msg = f"{remote} is not valid artifact repository for CMF.\n Reinitialize CMF."
            raise MsgFailure(msg_str=msg)


def add_parser(subparsers, parent_parser):
    PULL_HELP = "Pull artifacts from user configured repository."

    parser = subparsers.add_parser(
        "pull",
        parents=[parent_parser],
        description="Pull artifacts from user configured repository.",
        help=PULL_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        action="append",
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-f", "--file_name", action="append", help="Specify input metadata file name.", metavar="<file_name>"
    )

    parser.add_argument(
        "-a", "--artifact_name", action="append", help="Specify artifact name.", metavar="<artifact_name>"
    )

    parser.set_defaults(func=CmdArtifactPull)




###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import argparse
import textwrap
import readchar
import pandas as pd

from cmflib import cmfquery
from tabulate import tabulate
from cmflib.utils.helper_functions import display_table
from cmflib.cli.command import CmdBase
from cmflib.cmf_exception_handling import (
    PipelineNotFound,
    FileNotFound,
    DuplicateArgumentNotAllowed,
    MissingArgument,
    MsgSuccess,
    ExecutionUUIDNotFound
)

class CmdExecutionList(CmdBase):
    def run(self, live):
        cmd_args = {
            "file_name": self.args.file_name,
            "pipeline_name": self.args.pipeline_name,
            "execution_uuid": self.args.execution_uuid
        }
        for arg_name, arg_value in cmd_args.items():
            if arg_value:
                if arg_value[0] == "":
                    raise MissingArgument(arg_name)
                elif len(arg_value) > 1:
                    raise DuplicateArgumentNotAllowed(arg_name,("-"+arg_name[0]))
                
        current_directory = os.getcwd()
        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        else:
            mlmd_file_name = self.args.file_name[0].strip()
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
        
        current_directory = os.path.dirname(mlmd_file_name)
        if not os.path.exists(mlmd_file_name):  
            raise FileNotFound(mlmd_file_name, current_directory)

        # Creating cmfquery object.
        query = cmfquery.CmfQuery(mlmd_file_name)

        pipeline_name = self.args.pipeline_name[0]
        df = query.get_all_executions_in_pipeline(pipeline_name)

        # Check if the DataFrame is empty, indicating the pipeline name does not exist.
        if df.empty:    
            raise PipelineNotFound(pipeline_name)
        else:
            # Process execution ID if provided
            if not self.args.execution_uuid:         # If self.args.execution_uuid is None or an empty list ([]).
                pass
            else:
                df = df[df['Execution_uuid'].apply(lambda x: self.args.execution_uuid[0] in x.split(","))] # Used dataframe based on execution uuid
                if not df.empty:
                    # Rearranging columns: Start with fixed columns and appending the remaining columns.
                    updated_columns = ["id", "Context_Type", "Execution", "Execution_uuid", "name", "Pipeline_Type", "Git_Repo"] 
                    updated_columns += [ col for col in df.columns if col not in updated_columns]
                    
                    df = df[updated_columns]

                    # Drop columns that start with 'custom_properties_' and that contains NaN values
                    columns_to_drop = [col for col in df.columns if col.startswith('custom_properties_') and df[col].isna().any()]
                    df = df.drop(columns=columns_to_drop)

                    # Wrap text in object-type columns to a width of 30 characters.
                    for col in df.select_dtypes(include=['object']).columns:
                        df[col] = df[col].apply(lambda x: textwrap.fill(x, width=30) if isinstance(x, str) else x)
                    
                    # Set 'id' as the DataFrame index and transpose it for display horizontally.
                    df.set_index("id", inplace=True)
                    df = df.T.reset_index()
                    df.columns.values[0] = 'id'  # Rename the first column back to 'id'.

                    # Display the updated DataFrame as a formatted table.
                    table = tabulate(
                        df,
                        headers=df.columns,
                        tablefmt="grid",
                        showindex=False,
                    )
                    print(table)
                    print()
                    return MsgSuccess(msg_str = "Done.")
                return ExecutionUUIDNotFound(self.args.execution_uuid[0])
    
            display_table(df, ["id", "Context_Type", "Execution", "Execution_uuid", "name", "Pipeline_Type", "Git_Repo"])             
            return MsgSuccess(msg_str = "Done.")
    
    
def add_parser(subparsers, parent_parser):
    EXECUTION_LIST_HELP = "Displays executions from the input metadata file with a few properties in a 7-column table, limited to 20 records per page."

    parser = subparsers.add_parser(
        "list",
        parents=[parent_parser],
        description="Displays executions from the input metadata file with a few properties in a 7-column table, limited to 20 records per page.",
        help=EXECUTION_LIST_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    required_argumets = parser.add_argument_group("required arguments")

    required_argumets.add_argument(
        "-p", 
        "--pipeline_name",
        action="append", 
        required=True,
        help="Specify pipeline name.", 
        metavar="<pipeline_name>", 
    )

    parser.add_argument(
        "-f", 
        "--file_name", 
        action="append",
        help="Specify input metadata file name.",
        metavar="<file_name>",
    )

    parser.add_argument(
        "-e", 
        "--execution_uuid", 
        action="append",
        help="Specify the execution uuid to retrieve execution.",
        metavar="<exec_uuid>",
    )

    parser.set_defaults(func=CmdExecutionList)



###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import argparse

from cmflib.commands.execution import list
from cmflib.cli.utils import *

SUB_COMMANDS = [list]

# This parser adds positional argumets to the main parser
def add_parser(subparsers, parent_parser):
    LIST_HELP = "Display all executions with detailed information from the specified MLMD file."

    list_parser = subparsers.add_parser(
        "execution", 
        parents=[parent_parser],
        description="Display all executions with detailed information from the specified MLMD file.",
        help=LIST_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    list_subparsers = list_parser.add_subparsers(
        dest="cmd", help="Use `cmf execution CMD --help` for " "command-specific help."
    )

    fix_subparsers(list_subparsers)
    for cmd in SUB_COMMANDS:
        cmd.add_parser(list_subparsers, parent_parser)




###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
from typing import Optional, List

class CmfResponse(Exception):
    """
    Response and Exceptions raised by the CMF.
    CmfResponse includes two child classes,
        1. CmfSuccess
        2. CmfFailure
    On the basis of success and failure various child classes are created.

    Base class for all the cmf responses and  exceptions.
    """

    def __init__(self, return_code=None, status="failure", *args):
        self.return_code = return_code
        self.status = status
        super().__init__(*args)


class CmfFailure(CmfResponse):
    def __init__(self, return_code=None, *args):
        super().__init__(return_code, status="failure", *args)


# Subclass for Success Cases
class CmfSuccess(CmfResponse):
    def __init__(self, return_code=None, *args):
        super().__init__(return_code, status="success", *args)


"""CMF Success Class"""


class ExecutionsAlreadyExists(CmfSuccess):
    def __init__(self, return_code=201):
        super().__init__(return_code)

    def handle(self):
        return "INFO: Executions already exists."


class ObjectDownloadSuccess(CmfSuccess):
    def __init__(self, object_name, download_loc, return_code=202):
        self.object_name = object_name
        self.download_loc = download_loc
        super().__init__(return_code)

    def handle(self):
        return f"SUCCESS: Object {self.object_name} downloaded at {self.download_loc}."


class BatchDownloadSuccess(CmfSuccess):
    def __init__(self, files_downloaded, return_code=203):
        self.files_downloaded = files_downloaded
        super().__init__(return_code)

    def handle(self):
        return f"SUCCESS: Number of files downloaded = {self.files_downloaded }."


class MlmdFilePullSuccess(CmfSuccess):
    def __init__(self, full_path_to_dump, return_code=204):
        self.full_path_to_dump = full_path_to_dump
        super().__init__(return_code)

    def handle(self):
        return f"SUCCESS: {self.full_path_to_dump} is successfully pulled."


class MlmdFilePushSuccess(CmfSuccess):
    def __init__(self, file_name, return_code=205):
        self.file_name = file_name
        super().__init__(return_code)

    def handle(self):
        return f"SUCCESS: {self.file_name} is successfully pushed."


class TensorboardPushSuccess(CmfSuccess):
    def __init__(self, tensorboard_file_name: str = "All", return_code=206):
        self.tensorboard_file_name = tensorboard_file_name
        super().__init__(return_code)

    def handle(self):
        if self.tensorboard_file_name == "All":
            return f"SUCCESS: All tensorboard logs pushed successfully."
        return (
            f"tensorboard logs: file {self.tensorboard_file_name} pushed successfully."
        )


class CmfInitComplete(CmfSuccess):
    def __init__(self, return_code=207):
        super().__init__(return_code)

    def handle(self):
        return "SUCCESS: cmf init complete."


class CmfInitShow(CmfSuccess):
    def __init__(self, attr_str, return_code=208):
        self.attr_str = attr_str
        super().__init__(return_code)

    def handle(self):
        return f"{self.attr_str}"


class ArtifactPushSuccess(CmfSuccess):
    def __init__(self, message, return_code=209):
        self.message = message
        super().__init__(return_code)

    def handle(self):
        return self.message


class MetadataExportToJson(CmfSuccess):
    def __init__(self, full_path_to_dump, return_code=210):
        self.full_path_to_dump = full_path_to_dump
        super().__init__(return_code)

    def handle(self):
        return f"SUCCESS: metadata successfully exported in {self.full_path_to_dump}."


# This class is created for messages like "Done", "Records not found"
class MsgSuccess(CmfSuccess):
    def __init__(
        self,
        msg_str: Optional[str] = None,
        msg_list: Optional[List[str]] = None,
        return_code=211,
    ):
        self.msg_str = msg_str
        self.msg_list = msg_list
        super().__init__(return_code)

    def handle(self):
        if self.msg_list != None:
            return self.msg_list
        else:
            return self.msg_str


""" CMF FAILURE CLASSES"""


class PipelineNotFound(CmfFailure):
    def __init__(self, pipeline_name, return_code=101):
        self.pipeline_name = pipeline_name
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: Pipeline name {self.pipeline_name} doesn't exist."


class FileNotFound(CmfFailure):
    def __init__(self, file_name, directory, return_code=102):
        self.directory = directory
        self.file_name = file_name
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: File {self.file_name} doesn't exists in {self.directory} directory."


class BucketNotFound(CmfFailure):
    def __init__(self, bucket_name, return_code=103):
        self.bucket_name = bucket_name
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: Bucket {self.bucket_name} doesn't exist."


class ExecutionsNotFound(CmfFailure):
    def __init__(self, return_code=104):
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: Executions not found."


class ExecutionUUIDNotFound(CmfFailure):
    def __init__(self, exec_uuid, return_code=105):
        self.exec_uuid = exec_uuid
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: Execution uuid {self.exec_uuid} is not present in mlmd."


class ArtifactNotFound(CmfFailure):
    def __init__(self, artifact_name, return_code=106):
        self.artifact_name = artifact_name
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: Artifact {self.artifact_name} not found."


class ObjectDownloadFailure(CmfFailure):
    def __init__(self, object_name, return_code=107):
        self.object_name = object_name
        super().__init__(return_code)

    def handle(self):
        return f"Object {self.object_name} is not downloaded."


class BatchDownloadFailure(CmfFailure):
    def __init__(self, files_downloaded, Files_failed_to_download, return_code=108):
        self.files_downloaded = files_downloaded
        self.Files_failed_to_download = Files_failed_to_download
        super().__init__(return_code)

    def handle(self):
        return f"INFO: Number of files downloaded = {self.files_downloaded }. Files failed to download = {self.Files_failed_to_download}."


class Minios3ServerInactive(CmfFailure):
    def __init__(self, return_code=109):
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: MinioS3 server is not running!!!"


class CmfNotConfigured(CmfFailure):
    def __init__(self, message, return_code=110):
        self.message = message
        super().__init__(return_code)

    def handle(self):
        return self.message


class MlmdNotFoundOnServer(CmfFailure):
    def __init__(self, return_code=111):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: Metadata file not available on cmf-server."


class UpdateCmfVersion(CmfFailure):
    def __init__(self, return_code=112):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: You need to update cmf to the latest version. Unable to push metadata file."


class TensorboardPushFailure(CmfFailure):
    def __init__(self, tensorboard_file_name, response_text, return_code=113):
        self.tensorboard_file_name = tensorboard_file_name
        self.response_text = response_text
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: Failed to upload file {self.tensorboard_file_name}. Server response: {self.response_text}."


class Neo4jArgumentNotProvided(CmfFailure):
    def __init__(self, return_code=114):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: Provide user, password and uri for neo4j initialization."


class CmfInitFailed(CmfFailure):
    def __init__(self, return_code=115):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: cmf init failed."


class CmfServerNotAvailable(CmfFailure):
    def __init__(self, return_code=116):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: cmf-server is not available."


class InternalServerError(CmfFailure):
    def __init__(self, return_code=117):
        super().__init__(return_code)

    def handle(self):
        return "cmf-server error: The server encountered an unexpected error."


class MlmdFilePullFailure(CmfFailure):
    def __init__(self, return_code=118):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: Unable to pull metadata file."


class DirectoryNotfound(CmfFailure):
    def __init__(self, dir, return_code=119):
        self.dir = dir
        super().__init__(return_code)

    def handle(self):
        return f"ERROR: {self.dir} doesn't exists."


class FileNameNotfound(CmfFailure):
    def __init__(self, return_code=120):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: Provide path with file name."


class NoDataFoundOsdf(CmfFailure):
    def __init__(self, return_code=121):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: No data received from the server."


class InvalidTensorboardFilePath(CmfFailure):
    def __init__(self, return_code=122):
        super().__init__(return_code)

    def handle(self):
        return "ERROR: Invalid tensorboard logs path. Provide valid file/folder path for tensorboard logs!!"


class DuplicateArgumentNotAllowed(CmfFailure):
    def __init__(self, argument_name, argument_flag, return_code=123):
        self.argument_flag = argument_flag
        self.argument_name = argument_name
        super().__init__(return_code)

    def handle(self):
        return f"Error: You can only provide one {self.argument_name} using the {self.argument_flag} flag."


class MissingArgument(CmfFailure):
    def __init__(self, argument_name, return_code=124):
        self.argument_name = argument_name
        super().__init__(return_code)

    def handle(self):
        return f"Error: Missing {self.argument_name}"


class NoChangesMadeInfo(CmfFailure):
    def __init__(self, return_code=125):
        super().__init__(return_code)

    def handle(self):
        return "INFO: No changes made to the file. Operation aborted."


class MsgFailure(CmfFailure):
    def __init__(
        self,
        msg_str: Optional[str] = None,
        msg_list: Optional[List[str]] = None,
        return_code=126,
    ):
        self.msg_str = msg_str
        self.msg_list = msg_list
        super().__init__(return_code)

    def handle(self):
        if self.msg_list != None:
            return self.msg_list
        else:
            return self.msg_str



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import paramiko

class SSHremoteArtifacts:

    def __init__(self, dvc_config_op):
        self.user = dvc_config_op["remote.ssh-storage.user"]
        self.password = dvc_config_op["remote.ssh-storage.password"]


    def download_file(
        self,
        host: str,
        current_directory: str,
        object_name: str,
        download_loc: str,
    ):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(
            paramiko.AutoAddPolicy()
        )  # this can lead to man in the middle attack, need to find another solution
        ssh.connect(host, username=self.user, password=self.password)
        sftp = ssh.open_sftp()
        dir_path = ""
        # in case download_loc is absolute path like /home/user/test/data.xml.gz
        # we need to make this absolute path a relative one by removing first '/'
        if os.path.isabs(download_loc):
            download_loc = download_loc[1:]
        if "/" in download_loc:
            dir_path, _ = download_loc.rsplit("/", 1)
        if dir_path != "":
            # creates subfolders needed as per artifacts' folder structure
            os.makedirs(dir_path, mode=0o777, exist_ok=True) 

        abs_download_loc = os.path.abspath(os.path.join(current_directory, download_loc))
        try:
            local_file_size = os.stat(object_name).st_size  # Get the size of the local file being uploaded
            # The put() method returns an SFTPAttributes object, which contains metadata about the uploaded file.
            # Therefore, response should be typed as SFTPAttributes.
            response: paramiko.SFTPAttributes = sftp.put(object_name, abs_download_loc)
            # we can close sftp connection as we have already downloaded the file
            sftp.close()
            ssh.close()
            # After upload, check if the uploaded file size matches the local file size
            if response.st_size == local_file_size:
                return object_name, abs_download_loc, True
            return  object_name, abs_download_loc, False
        except Exception as e:
            # this exception is for function sftp.put()
            sftp.close()
            ssh.close()
            return  object_name, abs_download_loc, False


    def download_directory(
        self,
        host: str,
        current_directory: str,
        object_name: str,
        download_loc: str,
    ):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(
            paramiko.AutoAddPolicy()
        )  # this can lead to man in the middle attack, need to find another solution
        ssh.connect(host, username=self.user, password=self.password)
        sftp = ssh.open_sftp()
        dir_path = ""
        # in case download_loc is absolute path like home/user/test/data.xml.gz
        # we need to make this absolute path a relative one by removing first '/'
        if os.path.isabs(download_loc):
            download_loc = download_loc[1:]
        if "/" in download_loc:
            dir_path, _ = download_loc.rsplit("/", 1)
        if dir_path != "":
            # creates subfolders needed as per artifacts' folder structure
            os.makedirs(dir_path, mode=0o777, exist_ok=True) 

        abs_download_loc = os.path.abspath(os.path.join(current_directory, download_loc))
                                               
        """"
        if object_name ends with .dir - it is a directory.
        we download .dir object with 'temp_dir' and remove 
        this after all the files from this .dir object is downloaded.
        """
        # in case of .dir, abs_download_loc is a absolute path for a folder
        os.makedirs(abs_download_loc, mode=0o777, exist_ok=True)

        # download .dir object
        temp_dir = f"{abs_download_loc}/temp_dir"
        try:
            # The put() method returns an SFTPAttributes object, which contains metadata about the uploaded file.
            # Therefore, response should be typed as SFTPAttributes.
            response: paramiko.SFTPAttributes = sftp.put(object_name, temp_dir)
            with open(temp_dir, 'r') as file:
                tracked_files = eval(file.read())

            # removing temp_dir
            if os.path.exists(temp_dir):
                os.remove(temp_dir)
            """
            object_name =  /home/user/ssh-storage/files/md5/dd/2d792b7cf6efb02231f85c6147e403.dir
            contains the path of the .dir on the artifact repo
            we need to remove the hash of the .dir from the object_name
            which will leave us with the artifact repo path
            """
            
            repo_path = "/".join(object_name.split("/")[:-2])

            total_files_in_directory = 0
            files_downloaded = 0
            for file_info in tracked_files:
                total_files_in_directory += 1
                relpath = file_info['relpath']
                md5_val = file_info['md5']
                # download_loc =  /home/user/datatslice/example-get-started/test/artifacts/raw_data
                # md5_val = a237457aa730c396e5acdbc5a64c8453
                # we need a2/37457aa730c396e5acdbc5a64c8453
                formatted_md5 = md5_val[:2] + '/' + md5_val[2:]
                temp_download_loc = f"{abs_download_loc}/{relpath}"
                temp_object_name = f"{repo_path}/{formatted_md5}"
                try:
                    obj = sftp.put(object_name, temp_download_loc)
                    sftp.close()
                    ssh.close()
                    if obj:
                        files_downloaded += 1
                        print(f"object {temp_object_name} downloaded at {temp_download_loc}.")
                    else:
                        print(f"object {temp_object_name} is not downloaded.")
                except Exception as e:
                    sftp.close()
                    ssh.close()
                    print(f"object {temp_object_name} is not downloaded.")

            # total_files - files_downloaded gives us the number of files which are failed to download
            if (total_files_in_directory - files_downloaded) == 0:   
                return total_files_in_directory, files_downloaded, True
            return total_files_in_directory, files_downloaded, False  
        except Exception as e:
            sftp.close()
            ssh.close()
            print(f"object {object_name} is not downloaded.")
            # We usually don't count .dir as a file while counting total_files_in_directory.
            # However, here we failed to download the .dir folder itself. 
            # So we need to make, total_files_in_directory = 1
            total_files_in_directory = 1 
            return total_files_in_directory, files_downloaded, False



###
# Copyright (2024) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###


#!/usr/bin/env python3
import argparse
import os

from cmflib.cmf_exception_handling import CmfInitComplete, CmfInitFailed, Neo4jArgumentNotProvided
from cmflib.cli.command import CmdBase
from cmflib.dvc_wrapper import (
    git_quiet_init,
    git_checkout_new_branch,
    git_initial_commit,
    git_add_remote,
    dvc_quiet_init,
    dvc_add_remote_repo,
    dvc_add_attribute,
)
from cmflib.utils.cmf_config import CmfConfig
from cmflib.utils.helper_functions import is_git_repo
from cmflib.utils.helper_functions import generate_osdf_token

class CmdInitOSDFRemote(CmdBase):
    def run(self, live):
        # User can provide different name for cmf configuration file using CONFIG_FILE environment variable.
        # If CONFIG_FILE is not provided, default file name is .cmfconfig
        cmf_config = os.environ.get("CONFIG_FILE", ".cmfconfig")

        attr_dict = {}
        # cmf_server_url is default parameter for cmf init command 
        # if user does not provide cmf-server-url, default value is http://127.0.0.1:80
        attr_dict["server-url"] = self.args.cmf_server_url
        CmfConfig.write_config(cmf_config, "cmf", attr_dict)

        # read --neo4j details and add to the exsting file
        if self.args.neo4j_user and self.args.neo4j_password and self.args.neo4j_uri:
            attr_dict = {}
            attr_dict["user"] = self.args.neo4j_user
            attr_dict["password"] = self.args.neo4j_password
            attr_dict["uri"] = self.args.neo4j_uri
            CmfConfig.write_config(cmf_config, "neo4j", attr_dict, True)
        elif (
            not self.args.neo4j_user
            and not self.args.neo4j_password
            and not self.args.neo4j_uri
        ):
            pass
        else:
            raise Neo4jArgumentNotProvided
        output = is_git_repo()
        if not output:
            branch_name = "master"
            print("Starting git init.")
            git_quiet_init()
            git_checkout_new_branch(branch_name)
            git_initial_commit()
            git_add_remote(self.args.git_remote_url)
            print("git init complete.")

        print("Starting cmf init.")
        repo_type = "osdf"
        dvc_quiet_init()
        output = dvc_add_remote_repo(repo_type, self.args.path)
        if not output:
            raise CmfInitFailed
        print(output)
        #dvc_add_attribute(repo_type, "key_id", self.args.key_id)
        #dvc_add_attribute(repo_type, "key_path", self.args.key_path)
        #dvc_add_attribute(repo_type, "key_issuer", self.args.key_issuer)
        #Writing to an OSDF Remote is based on SSH Remote. With few additions
        #In addition to URL (including FQDN, port, path), we need to provide 
        #method=PUT, ssl_verify=false, ask_password=false, auth=custom, custom_auth-header='Authorization'
        #password='Bearer + dynamically generated scitoken' (This token has a timeout of 15 mins so must be generated right before any push/pull) 
        dvc_add_attribute(repo_type,"method", "PUT")
        dvc_add_attribute(repo_type,"ssl_verify", "false")
        dvc_add_attribute(repo_type,"ask_password", "false")
        dvc_add_attribute(repo_type,"auth", "custom")
        dvc_add_attribute(repo_type,"custom_auth_header", "Authorization")
        dynamic_password = generate_osdf_token(self.args.key_id,self.args.key_path,self.args.key_issuer)
        dvc_add_attribute(repo_type,"password",dynamic_password)

        attr_dict = {}
        attr_dict["path"] = self.args.path
        attr_dict["cache"] = self.args.cache
        attr_dict["key_id"] = self.args.key_id
        attr_dict["key_path"] = self.args.key_path
        attr_dict["key_issuer"] = self.args.key_issuer
        CmfConfig.write_config(cmf_config, "osdf", attr_dict, True)

        return CmfInitComplete


def add_parser(subparsers, parent_parser):
    HELP = "Initialises remote OSDF directory as artifact repository."

    parser = subparsers.add_parser(
        "osdfremote",
        parents=[parent_parser],
        description="This command initialises remote OSDF directory as artifact repository for CMF.",
        help=HELP,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "--path",
        required=True,
        help="Specify FQDN for OSDF directory path including port and path",
        metavar="<path>",
        default=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--cache",
        help="Specify FQDN for OSDF cache path including port and path. For Ex. https://osdf-director.osg-htc.org/nrp/fdp/",
        metavar="<cache>",
        #default="https://osdf-director.osg-htc.org/nrp/fdp/",
        default="",
    )

    required_arguments.add_argument(
        "--key-id",
        required=True,
        help="Specify key_id for provided private key. eg. b2d3",
        metavar="<key_id>",
        default=argparse.SUPPRESS,
    )

    required_arguments.add_argument(
        "--key-path",
        required=True,
        help="Specify path for private key on local filesystem. eg. ~/.ssh/XXX.pem",
        metavar="<key_path>",
        default=argparse.SUPPRESS,
    )

    required_arguments.add_argument(
        "--key-issuer",
        required=True,
        help="Specify URL for Key Issuer. eg. https://t.nationalresearchplatform.org/XXX",
        metavar="<key_issuer>",
        default=argparse.SUPPRESS,
    )

    required_arguments.add_argument(
        "--git-remote-url",
        required=True,
        help="Specify git repo url. eg: https://github.com/XXX/example.git",
        metavar="<git_remote_url>",
        default=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--cmf-server-url",
        help="Specify cmf-server URL.",
        metavar="<cmf_server_url>",
        default="http://127.0.0.1:8080",
    )

    parser.add_argument(
        "--neo4j-user",
        help="Specify neo4j user.",
        metavar="<neo4j_user>",
        # default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--neo4j-password",
        help="Specify neo4j password.",
        metavar="<neo4j_password>",
        # default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--neo4j-uri",
        help="Specify neo4j uri.eg bolt://localhost:7687",
        metavar="<neo4j_uri>",
        # default=argparse.SUPPRESS,
    )

    parser.set_defaults(func=CmdInitOSDFRemote)



###
# Copyright (2022) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import os
import json
import argparse

from cmflib import cmfquery
from cmflib.cli.command import CmdBase
from cmflib.cmf_exception_handling import (
    PipelineNotFound,
    FileNotFound,
    DuplicateArgumentNotAllowed,
    MissingArgument,
    NoChangesMadeInfo,
    MetadataExportToJson,
    DirectoryNotfound,
    MsgFailure
)

# This class export local mlmd data to a json file
class CmdMetadataExport(CmdBase):
    def create_full_path(self, current_directory: str, json_file_name: str) -> str:
        if not os.path.isdir(json_file_name):
            temp = os.path.dirname(json_file_name)
            current_directory = './'
            if temp != "":
                current_directory = temp
            if os.path.exists(current_directory):
                full_path_to_dump  = json_file_name
                return full_path_to_dump
            else:
                raise DirectoryNotfound(current_directory)
        else:
            raise MsgFailure(msg_str = "Provide path with file name.")
        
    def run(self, live):
        cmd_args = {
            "file_name": self.args.file_name,
            "pipeline_name": self.args.pipeline_name,
            "json_file_name": self.args.json_file_name
        }  
        for arg_name, arg_value in cmd_args.items():
            if arg_value:
                if arg_value[0] == "":
                    raise MissingArgument(arg_name)
                elif len(arg_value) > 1:
                    raise DuplicateArgumentNotAllowed(arg_name,("-"+arg_name[0]))
         
        current_directory = os.getcwd()
        full_path_to_dump = ""

        if not self.args.file_name:         # If self.args.file_name is None or an empty list ([]). 
            mlmd_file_name = "./mlmd"       # Default path for mlmd file name.
        else:
            mlmd_file_name = self.args.file_name[0].strip() # Removing starting and ending whitespaces.
            if mlmd_file_name == "mlmd":
                mlmd_file_name = "./mlmd"
        
        current_directory = os.path.dirname(mlmd_file_name)
        if not os.path.exists(mlmd_file_name): 
            raise FileNotFound(mlmd_file_name, current_directory)
        
        # Initialising cmfquery class.
        query = cmfquery.CmfQuery(mlmd_file_name)

        pipeline_name = self.args.pipeline_name[0]

        if pipeline_name in query.get_pipeline_names():
            if not self.args.json_file_name:         # If self.args.json_file_name is None or an empty list ([]). 
                json_file_name = self.args.json_file_name
            else:
                json_file_name = self.args.json_file_name[0].strip()

            # Setting directory where mlmd file will be dumped.
            if json_file_name:
                if not json_file_name.endswith(".json"):
                    json_file_name = json_file_name+".json" # Added .json extention to json file name.
                if os.path.exists(json_file_name):
                    live.stop()
                    userRespone = input("File name already exists do you want to continue press yes/no: ")
                    if userRespone.lower() == "yes":    # Overwrite file.
                        live.start()
                        full_path_to_dump = self.create_full_path(current_directory, json_file_name)
                    else: 
                        raise NoChangesMadeInfo()
                else:  
                    full_path_to_dump = self.create_full_path(current_directory, json_file_name)
            else: 
                # Checking whether a json file exists in the directory based on pipeline name.
                if os.path.exists(f"{pipeline_name}.json"):
                    live.stop()
                    userRespone = input("File name already exists do you want to continue press yes/no: ")
                    if userRespone.lower() == "yes":
                        live.start()
                        full_path_to_dump = os.getcwd() + f"/{pipeline_name}.json"
                    else:
                        raise NoChangesMadeInfo()
                else:  
                    full_path_to_dump = os.getcwd() + f"/{pipeline_name}.json"

            # Pulling data from local mlmd file.
            json_payload = query.dumptojson(pipeline_name, None)

            # Write metadata into json file.
            with open(full_path_to_dump, 'w') as f:
                f.write(json.dumps(json.loads(json_payload),indent=2))
                return MetadataExportToJson(full_path_to_dump)
        else:
            raise PipelineNotFound(pipeline_name)
            


def add_parser(subparsers, parent_parser):
    PULL_HELP = "Export local metadata's metadata in json format to a json file. "

    parser = subparsers.add_parser(
        "export",
        parents=[parent_parser],
        description="Export local metadata's metadata in json format to a json file.",
        help=PULL_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        action="append",
        required=True,
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-j",
        "--json_file_name",
        action="append",
        help="Specify output json file name with full path.",
        metavar="<json_file_name>",
    )

    parser.add_argument(
        "-f", 
        "--file_name", 
        action="append",
        help="Specify the absolute or relative path for the input metadata file.", 
        metavar="<file_name>",
    )

    parser.set_defaults(func=CmdMetadataExport)
